class Node {
  String label;
  ArrayList<Node> edges = new ArrayList<Node>();
  boolean isRoot = false;

  PVector position = new PVector();
  PVector velocity = new PVector();
  PVector acceleration = new PVector();

  float maxforce = 100.0;
  float maxspeed = 1.0;

  float desiredseparation = 100.0f;
  float neighbordist = 50000;

  public Node(String label, boolean isRoot) {
    this.label = label;

    this.position.x = random(0, width);
    this.position.y = random(0, width);

    if (isRoot) {
      this.position.x = 0;
      this.position.y = 0;
    }

    this.isRoot = isRoot;
  }

  public Node(String label) {
    this(label, false);
  }

  public void addEdges(Node ... nodes) {
    for (Node node : nodes) {
      edges.add(node);
    }
  }

  void iterate(ArrayList<Node> nodes) {
    if(edges.size() < 1)
      return;
    
    if (!isRoot) {
      // forces
      PVector sep = separate(nodes);
      PVector coh = cohesion(nodes);
      PVector cohEdges = cohesion(edges);

      sep.mult(1.6);
      coh.mult(1.0);
      cohEdges.mult(1.5);

      acceleration.add(sep);
      acceleration.add(coh);
      acceleration.add(cohEdges);

      // move
      velocity.add(acceleration);
      velocity.limit(maxspeed);
      position.add(velocity);
      acceleration.mult(0);
    }
  }

  void render(ArrayList<Node> nodes) {
    // visual
    stroke(#003f91, 100);
    fill(#5da9e9, 100);

    if (isRoot) {
      stroke(#FC8F7E);
      fill(#FE9095, 100);
    }

    circle(position.x, position.y, edges.size());

    fill(#003f91);
    textSize(8);

    if (isRoot) {
      fill(#FE9095);
    }

    text(label, position.x + 8, position.y + 8);

    for (Node edge : edges) {
      strokeWeight(0.5);
      line(position.x, position.y, edge.position.x, edge.position.y);
    }
  }

  PVector separate (ArrayList<Node> nodes) {
    PVector steer = new PVector(0, 0, 0);
    int count = 0;

    for (Node other : nodes) {
      float d = PVector.dist(position, other.position);

      if ((d > 0) && (d < desiredseparation)) {
        PVector diff = PVector.sub(position, other.position);
        diff.normalize();
        diff.div(d);
        steer.add(diff);
        count++;
      }
    }

    if (count > 0) {
      steer.div((float)count);
    }

    if (steer.mag() > 0) {
      steer.normalize();
      steer.mult(maxspeed);
      steer.sub(velocity);
      steer.limit(maxforce);
    }
    return steer;
  }

  PVector cohesion (ArrayList<Node> nodes) {
    PVector sum = new PVector(0, 0);
    int count = 0;
    for (Node other : nodes) {
      float d = PVector.dist(position, other.position);
      if ((d > 0) && (d < neighbordist)) {
        sum.add(other.position);
        count++;
      }
    }
    if (count > 0) {
      sum.div(count);
      return seek(sum);
    } else {
      return new PVector(0, 0);
    }
  }

  PVector seek(PVector target) {
    PVector desired = PVector.sub(target, position);
    desired.setMag(maxspeed);

    PVector steer = PVector.sub(desired, velocity);
    steer.limit(maxforce);
    return steer;
  }
}
